/* -*- mode: c++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/*
 Copyright (C) 2012 Klaus Spanderen

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.

 compile:
nvcc -I/home/klaus/workspace/QuantLib -I. -arch=compute_20 -O3 -c gpubrownianbridge.cu
*/

#include <gpubrownianbridge.hpp>

#include <thrust/copy.h>
#include <thrust/for_each.h>
#include <thrust/iterator/counting_iterator.h>


#define CURAND_CALL(x) do { if((x) != CURAND_STATUS_SUCCESS) { \
    char s[255];                                               \
    sprintf(s, "Error %i at %s:%d\n", x, __FILE__,__LINE__);   \
    QL_FAIL(s);                                                \
}} while(0)

#define BUFFER_SIZE 32u

namespace QuantLib {

    struct bridge_functor {
        bridge_functor(uint32_t nPaths, uint32_t factors, uint32_t steps, 
                       const uint32_t* const orderedIndices,
                       const uint32_t* const bridgeIdx,
                       const uint32_t* const leftIdx,
                       const uint32_t* const rightIdx,
                       const real* const leftWeight,
                       const real* const rightWeight,
                       const real* const stdDev,
                       const real* const input,
                       real* const output)
        : nPaths_(nPaths), factors_(factors), steps_(steps),
          orderedIndices_(orderedIndices),
          bridgeIdx_(bridgeIdx), leftIdx_(leftIdx), rightIdx_(rightIdx),
          leftWeight_(leftWeight), rightWeight_(rightWeight), stdDev_(stdDev),
          input_(input), output_(output) {}

        __device__
        void operator()(const uint32_t& idx) const {
            const uint32_t bIdx =  threadIdx.x + threadIdx.y*blockDim.x
                                 + threadIdx.z*blockDim.x*blockDim.y;
            const uint32_t copySize 
                = min(blockDim.x*blockDim.y*blockDim.z, BUFFER_SIZE);

            const uint32_t iPath   = idx / factors_;
            const uint32_t iFactor = idx % factors_;
            const uint32_t stepJump = nPaths_*factors_;
            
            real* const out = output_ + iPath + iFactor*nPaths_;
            const real* const in  = input_ + iPath;
            const uint32_t* const orderedIdx = orderedIndices_+iFactor*steps_;

			__shared__ float sLeftWeight[BUFFER_SIZE];
			__shared__ float sRightWeight[BUFFER_SIZE];
			__shared__ float sStdDev[BUFFER_SIZE];
			__shared__ uint32_t sLeftIndex[BUFFER_SIZE];
			__shared__ uint32_t sRightIndex[BUFFER_SIZE];
			__shared__ uint32_t sBridgeIndex[BUFFER_SIZE];

            out[(steps_-1u)*stepJump] = stdDev_[0]*in[orderedIdx[0]*nPaths_];

            uint32_t ii = copySize-1u;

            for (uint32_t i=1u; i < steps_; ++i) {
                if (++ii == copySize) {
                    __syncthreads();

                    const uint32_t ibIdx = i+bIdx;
                    if (bIdx < copySize && ibIdx < steps_) {
                        sRightIndex [bIdx] = rightIdx_   [ibIdx];
                        sLeftIndex  [bIdx] = leftIdx_    [ibIdx];
                        sBridgeIndex[bIdx] = bridgeIdx_  [ibIdx];
                        sLeftWeight [bIdx] = leftWeight_ [ibIdx];
                        sRightWeight[bIdx] = rightWeight_[ibIdx];
                        sStdDev     [bIdx] = stdDev_     [ibIdx];
                    }
                    ii = 0u;

                    __syncthreads();
                }

                const uint32_t j = sLeftIndex[ii];
                const uint32_t k = sRightIndex[ii];
                const uint32_t l = sBridgeIndex[ii];

                if (j != 0u) {
                    out[l*stepJump] =
                        sLeftWeight[ii] * out[(j-1)*stepJump] +
                        sRightWeight[ii] * out[k*stepJump]   +
                        sStdDev[ii] * in[orderedIdx[i]*nPaths_];
                } else {
                    out[l*stepJump] =
                        sRightWeight[ii] * out[k*stepJump]   +
                        sStdDev[ii] * in[orderedIdx[i]*nPaths_];
                }
            }
            for (Size i=steps_-1; i>=1; --i) {
                out[i*stepJump] -= out[(i-1)*stepJump];
            }
        }

      private:
        const uint32_t nPaths_, factors_, steps_;
        const uint32_t* const orderedIndices_;
        const uint32_t* const bridgeIdx_;
        const uint32_t* const leftIdx_;
        const uint32_t* const rightIdx_;
        const real* const leftWeight_;
        const real* const rightWeight_;
        const real* const stdDev_;
        const real* const input_;
        real* const output_;
    };

    GPUBrownianBridge::GPUBrownianBridge(
        Size factors, Size steps, 
        SobolBrownianGenerator::Ordering ordering,
        unsigned long long seed,
        unsigned long long offset, 
        curandRngType rngType)
    : factors_    (factors),
      steps_      (steps),
      bridge_     (new BrownianBridge(steps)),
      bridgeIdx_  (bridge_->bridgeIndex()),
      leftIdx_    (bridge_->leftIndex()),
      rightIdx_   (bridge_->rightIndex()),
      leftWeight_ (bridge_->leftWeight()),
      rightWeight_(bridge_->rightWeight()),
      stdDev_     (bridge_->stdDeviation()),
      orderedIndices_(steps*factors),
      gen_(0) {

        SobolBrownianGenerator sobolGen(factors, steps, ordering, seed);
        const std::vector<std::vector<Size> >& orderedIndices = 
            sobolGen.orderedIndices();
        
        for (Size i=0; i < factors; ++i) {
            thrust::copy(orderedIndices[i].begin(), orderedIndices[i].end(), 
                         orderedIndices_.begin()+i*steps);
        }

        CURAND_CALL(curandCreateGenerator(&gen_, rngType));
        CURAND_CALL(curandSetGeneratorOffset(gen_, offset + 1ull));

        switch(rngType) {
          case CURAND_RNG_PSEUDO_DEFAULT:
          case CURAND_RNG_PSEUDO_XORWOW:
            CURAND_CALL(curandSetPseudoRandomGeneratorSeed(gen_, seed));
            break;
          case CURAND_RNG_QUASI_DEFAULT:
          case CURAND_RNG_QUASI_SOBOL32:
          case CURAND_RNG_QUASI_SCRAMBLED_SOBOL32:
          case CURAND_RNG_QUASI_SOBOL64:
          case CURAND_RNG_QUASI_SCRAMBLED_SOBOL64:
            CURAND_CALL(curandSetQuasiRandomGeneratorDimensions(
                gen_, steps*factors));
            break;
          default:
            QL_FAIL("unknown curandRngType given");
        }
    }


    GPUBrownianBridge::~GPUBrownianBridge() {
        CURAND_CALL(curandDestroyGenerator(gen_));
    }


    namespace {
        curandStatus_t curandNormalGenerator(curandGenerator_t generator, 
                                             float* outputPtr, size_t n, 
                                             float mean, float stddev) {
            return curandGenerateNormal(generator, outputPtr, n, mean, stddev);
        }

        curandStatus_t curandNormalGenerator(curandGenerator_t generator, 
                                             double* outputPtr, size_t n, 
                                             float mean, float stddev) {
            return curandGenerateNormalDouble(generator, 
                                              outputPtr, n, mean, stddev);
        }
    }

    void GPUBrownianBridge::transform(
        thrust::device_vector<real>::iterator begin,
        thrust::device_vector<real>::iterator end,
        thrust::device_vector<real>::iterator output) {

        const uint32_t n = thrust::distance(begin, end);
        const uint32_t length = steps_*factors_;

        QL_REQUIRE(!(n % length), "sequence lengths must be a "
                   "multiple of the size of the brownian bridge");

        const uint32_t nPaths = n/length;

        CURAND_CALL(curandNormalGenerator(
            gen_,  thrust::raw_pointer_cast(begin.base()), n ,0.0, 1.0));

        bridge_functor f(nPaths, factors_, steps_,
                         thrust::raw_pointer_cast(&orderedIndices_[0]),
                         thrust::raw_pointer_cast(&bridgeIdx_[0]),
                         thrust::raw_pointer_cast(&leftIdx_[0]),
                         thrust::raw_pointer_cast(&rightIdx_[0]),
                         thrust::raw_pointer_cast(&leftWeight_[0]),
                         thrust::raw_pointer_cast(&rightWeight_[0]),
                         thrust::raw_pointer_cast(&stdDev_[0]),
                         thrust::raw_pointer_cast(begin.base()),
                         thrust::raw_pointer_cast(output.base()));

        const thrust::counting_iterator<uint32_t> start(0u);
        thrust::for_each(start, start + nPaths*factors_, f);
    }  
}
